from collections import namedtuple
import torch.nn as nn
from .utils import _isinstance
from .fold import isfoldmodule

# global config for tqt
Fconfig = namedtuple('f_config', ['weight', 'bias', 'acti'],
                     defaults=[8, 8, 8])


def list_to_fconfig(lst, default=[8, 8, 8]):
    config = []
    for i in lst:
        if i == []:
            config.append(Fconfig(default))
        elif len(i) == len(default):
            config.append(Fconfig(i))
        else:
            raise Exception('the config length is incorrect')


def get_all_layers(net, al=[]):
    '''
    get all layers of (fusedmodule/nn.module) for config 
    '''
    for i in net._modules.keys():
        if list(net._modules[i]._modules) == []:
            if not isinstance(net._modules[i], nn.Identity):
                al.append(net._modules[i])
        elif isfoldmodule(net._modules[i]):
            al.append(net._modules[i])
        else:
            al = get_all_layers(net._modules[i], al)
    return al


def layers_config(net, lst):
    '''
    net is a nn.module with folded modules
    lst is a list of Fconfig, indicating the w/b/a of each layer
    '''
    list_flag = 0
    module_flag = 0
    key_len = len(get_all_layers(net))
    key = list(get_all_layers(net))
    list_len = len(lst)
    
    if list_len != key_len:
        raise Exception('config length not match')
    
    for mod, cfg in zip(key, lst):
        if hasattr(mod, 'weight_bit_width'):
            mod.weight_bit_width = cfg.weight
        if hasattr(mod, 'bias_bit_width'):
            mod.bias_bit_width = cfg.bias
        if hasattr(mod, 'acti_bit_width'):
            mod.acti_bit_width = cfg.acti


import torch
import torch.nn as nn
from collections import OrderedDict
from .convbnact import Conv2dBNReLU
from .convbn import Conv2dBN
from .sabn import SA2dBN
from .sabnact import SA2dBNReLU
from ..function.extra import SEConv2d, Adder2d, adder2d_function


def fold_CBR(conv, bn, relu):
    folded = Conv2dBNReLU(conv, bn, relu)
    conv = folded
    bn = nn.Identity()
    relu = nn.Identity()
    return conv, bn, relu


def fold_CB(conv, bn):
    folded = Conv2dBN(conv, bn)
    conv = folded
    bn = nn.Identity()
    return conv, bn


def fold_SABR(sa, bn, relu):
    folded = SA2dBNReLU(sa, bn, relu)
    sa = folded
    bn = nn.Identity()
    relu = nn.Identity()
    return sa, bn, relu


def fold_SAB(sa, bn):
    folded = SA2dBN(sa, bn)
    sa = folded
    bn = nn.Identity()
    return sa, bn

def fold_the_convnet(net):
    key = list(net._modules.keys())
    keylen = len(key)
    # for convbn
    flag = 0
    while flag < keylen:
        if list(net._modules[key[flag]]._modules.keys()) != []:
            fold_the_network(net._modules[key[flag]])
        else:
            if isinstance(net._modules[key[flag]],
                          nn.Conv2d) and flag + 1 < keylen:
                if isinstance(net._modules[key[flag + 1]],
                              nn.BatchNorm2d) and flag + 2 < keylen:
                    if isinstance(net._modules[key[flag + 2]],
                                  nn.ReLU) or isinstance(
                                      net._modules[key[flag + 2]], nn.ReLU6):
                        net._modules[key[flag]], net._modules[key[
                            flag + 1]], net._modules[key[flag + 2]] = fold_CBR(
                                net._modules[key[flag]],
                                net._modules[key[flag + 1]],
                                net._modules[key[flag + 2]])
                        flag += 2
                    else:
                        net._modules[key[flag]], net._modules[key[
                            flag + 1]] = fold_CB(net._modules[key[flag]],
                                                 net._modules[key[flag + 1]])
                        flag += 1
                elif isinstance(net._modules[key[flag + 1]], nn.BatchNorm2d):
                    net._modules[key[flag]], net._modules[key[
                        flag + 1]] = fold_CB(net._modules[key[flag]],
                                             net._modules[key[flag + 1]])
                    flag += 1
        flag += 1

def fold_the_addernet(net):
    key = list(net._modules.keys())
    keylen = len(key)
    flag = 0
    while flag < keylen:
        # if list(net._modules[key[flag]]._modules.keys()) != []:
        fold_the_network(net._modules[key[flag]])
        # else:
        if isinstance(net._modules[key[flag]],
                      nn.Sequential) and flag + 1 < keylen:
            if isinstance(net._modules[key[flag]][0], SEConv2d) and isinstance(
                    net._modules[key[flag]][1], Adder2d):
                if isinstance(net._modules[key[flag + 1]],
                              nn.BatchNorm2d) and flag + 2 < keylen:
                    if isinstance(net._modules[key[flag + 2]],
                                  nn.ReLU) or isinstance(
                                      net._modules[key[flag + 2]], nn.ReLU6):
                        net._modules[key[flag]], net._modules[
                            key[flag +
                                1]], net._modules[key[flag + 2]] = fold_SABR(
                                    net._modules[key[flag]],
                                    net._modules[key[flag + 1]],
                                    net._modules[key[flag + 2]])
                        flag += 2
                    else:
                        net._modules[key[flag]], net._modules[key[
                            flag + 1]] = fold_SAB(net._modules[key[flag]],
                                                  net._modules[key[flag + 1]])
                        flag += 1
                elif isinstance(net._modules[key[flag + 1]], nn.BatchNorm2d):
                    net._modules[key[flag]], net._modules[key[
                        flag + 1]] = fold_SAB(net._modules[key[flag]],
                                              net._modules[key[flag + 1]])
                    flag += 1
        flag += 1

def fold_the_network(net, fold_type='conv'):
    if fold_type == 'conv':
        fold_the_convnet(net) 
    elif fold_type == 'add':
        fold_the_addernet(net)
    else:
        raise NotImplementedError()

def make_the_shortcut_share(net, show=False):
    keys = list(net._modules.keys())
    if hasattr(net, 'share_path'):
        getattr(net, 'share_path')(show=show)
    for k in keys:
        make_the_shortcut_share(net._modules[k], show=show)



import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.nn.init as init
from torch.nn.modules.utils import _pair


class SEConv2d(nn.Module):
    def __init__(self,
                 in_channels,
                 out_channels,
                 kernel_size,
                 stride=1,
                 padding=0,
                 dilation=1,
                 groups=1,
                 bias=False,
                 size_splits=64,
                 threshold=5e-3,
                 sign_threshold=0.5,
                 distribution='uniform'):
        super(SEConv2d, self).__init__()
        if in_channels % groups != 0:
            raise ValueError('in_channels must be divisible by groups')
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.kernel_size = _pair(kernel_size)
        self.stride = _pair(stride)
        self.padding = _pair(padding)
        self.dilation = _pair(dilation)
        self.groups = groups
        if bias:
            self.bias = nn.Parameter(torch.Tensor(out_channels))
        else:
            self.register_parameter('bias', None)

        self.weight = torch.nn.Parameter(
            nn.init.normal_(
                torch.randn(self.out_channels, self.in_channels, kernel_size,
                            kernel_size)))

    def forward(self, input):
        weight = self.weight.detach()
        output = F.conv2d(input, weight, self.bias, self.stride, self.padding,
                          self.dilation, self.groups)

        return output



import torch
import torch.nn as nn
from ..function import qsigned


class ShareQuant(nn.Module):
    def __init__(self, acti_bit_width=8, retrain=True, quant=False):
        super().__init__()
        self.acti_bit_width = acti_bit_width
        self.acti_log2_t = nn.Parameter(torch.Tensor(1))
        self.retrain = retrain
        self.quant = quant

    def share(self, shared_module):
        self.acti_log2_t.data = max(shared_module.acti_log2_t.data,
                                    self.acti_log2_t.data)

    def static(self):
        self.retrain = False
        self.acti_log2_t.requires_grad_(False)

    def quantilize(self):
        self.quant = True
        self.acti_log2_t.requires_grad_(True)

    def floatilize(self):
        self.quant = False
        self.acti_log2_t.requires_grad_(False)

    def share_forward(self, input):
        return qsigned(input, self.acti_log2_t, self.acti_bit_width)

    def share_forward_unquant(self, input):
        return input

    def forward(self, input):
        return self.share_forward(
            input) if self.quant else self.share_forward_unquant(input)



import torch
import torch.nn as nn
from ..function import qsigned, qunsigned
from ..function.extra import SEConv2d, Adder2d, adder2d_function
from .foldmodule import _FoldModule
from .utils import isShiftAdder


class SA2dBNReLU(_FoldModule):
    def __init__(self, shiftadder, bn, relu):
        super().__init__()
        if isShiftAdder(shiftadder) and isinstance(
                bn, nn.BatchNorm2d) and (isinstance(relu, nn.ReLU)
                                         or isinstance(relu, nn.ReLU6)):
            self.shift = shiftadder[0]
            self.adder = shiftadder[1]
            self.bn = bn
            self.relu = nn.ReLU() if isinstance(relu, nn.ReLU) else nn.ReLU6()
            self.bn_weight(bn)
            self.adder_weight(shiftadder[1])
            self.acti_bit_width = relu.acti_bit_width
            self.acti_log2_t = relu.acti_log2_t
        self.bn_freezing = False
        self.quant = False

    def adder_weight(self, adder):
        self.adder_weight_bit_width = adder.weight_bit_width
        self.adder_weight_log2_t = adder.weight.abs().max().detach().data.log2(
        )

    def bn_weight(self, bn):
        bn_var = bn.running_var.detach().clone().data.reshape(-1, 1, 1, 1)
        bn_mean = bn.running_mean.detach().clone().data.reshape(-1, 1, 1, 1)
        bn_weight = bn.weight.detach().clone().data.reshape(-1, 1, 1, 1)
        bn_bias = bn.bias.detach().clone().data.reshape(-1, 1, 1, 1)
        bn_weight = bn_weight / (bn_var + bn.eps).sqrt()
        bn_bias = bn_weight * (-bn_mean) / (bn_var + bn.eps).sqrt() + bn_bias
        self.bn_weight_bit_width = bn.weight_bit_width
        self.bn_weight_log2_t = torch.nn.Parameter(
            bn_weight.abs().max().detach().data.log2())
        self.bn_bias_bit_width = bn.bias_bit_width
        self.bn_bias_log2_t = torch.nn.Parameter(
            bn_bias.abs().max().detach().data.log2())

    def bn_freeze(self, mode=True):
        self.bn_freezing = mode

    def quantilize(self):
        self.quant = True
        self.bn_weight_log2_t.requires_grad = True
        self.adder_weight_log2_t.requires_grad = True
        self.bn_bias_log2_t.requires_grad = True
        self.acti_log2_t.requires_grad = True

    def floatilize(self):
        self.quant = False
        self.bn_weight_log2_t.requires_grad = False
        self.adder_weight_log2_t.requires_grad = False
        self.bn_bias_log2_t.requires_grad = False
        self.acti_log2_t.requires_grad = False

    def forward(self, input):
        if self.bn_freezing:
            bn_var = self.bn.running_var.detach().clone().data.reshape(
                -1, 1, 1, 1)
            bn_mean = self.bn.running_mean.detach().clone().data.reshape(
                -1, 1, 1, 1)
            bn_weight = self.bn.weight.detach().clone().data.reshape(
                -1, 1, 1, 1)
            bn_bias = self.bn.bias.detach().clone().data.reshape(-1, 1, 1, 1)
        else:
            bn_var = self.bn.running_var.reshape(-1, 1, 1, 1)
            bn_mean = self.bn.running_mean.reshape(-1, 1, 1, 1)
            bn_weight = self.bn.weight.reshape(-1, 1, 1, 1)
            bn_bias = self.bn.bias.reshape(-1, 1, 1, 1)
        bn_weight = bn_weight / (bn_var + self.bn.eps).sqrt()
        bn_bias = bn_weight * (-bn_mean) / (bn_var +
                                            self.bn.eps).sqrt() + bn_bias
        if self.quant and self.bn_freezing:
            adder_weight = qsigned(self.adder.weight, self.adder_weight_log2_t,
                                   self.adder_weight_bit_width)
            bn_weight = qsigned(bn_weight, self.bn_weight_log2_t,
                                self.bn_weight_bit_width)
            bn_bias = qsigned(bn_bias, self.bn_bias_log2_t,
                              self.bn_bias_bit_width)
            inter = self.shift(input)
            inter = adder2d_function(inter,
                                     adder_weight,
                                     stride=self.adder.stride,
                                     padding=self.adder.padding)
            inter = bn_weight * inter + bn_bias
        elif self.quant and self.bn_freezing == False:
            adder_weight = qsigned(self.adder.weight, self.adder_weight_log2_t,
                                   self.adder_weight_bit_width)
            inter = self.shift(input)
            inter = adder2d_function(inter,
                                     adder_weight,
                                     stride=self.adder.stride,
                                     padding=self.adder.padding)
            inter = self.bn(inter)
        else:
            inter = self.shift(input)
            inter = adder2d_function(inter, self.adder.weight,
                                     self.adder.stride, self.adder.padding)
            inter = self.bn(inter)

        inter = self.relu(inter)

        if self.quant:
            inter = qunsigned(inter, self.acti_log2_t, self.acti_bit_width)
        return inter



from .convbnact import Conv2dBNReLU
from .convbn import Conv2dBN
from .sharequant import ShareQuant

from .sabn import SA2dBN
from .sabnact import SA2dBNReLU

from .fold_op import fold_CBR, fold_the_network, make_the_shortcut_share
from .foldmodule import isfoldmodule